from enum import Flag, auto
from os import path
from json import load
from math import sqrt, floor
from PIL import Image

class DataSource(Flag):
    """Flag to indicate origin dataset."""
    CHOLEC = auto()
    ROBUST = auto()
    BOTH = CHOLEC | ROBUST

class AnnotationType(Flag):
    """Flag to indicate type of annotation."""
    AREA = auto()
    MASK = auto()
    BOTH = AREA | MASK

class ECADataset():
    """
    Dataloader for the ECA dataset.

    Args:
        data_directory: root directory of the ECA data.
        data_source: flag denoting the origin dataset(s) samples will be taken from.
        annotation_type: flag denoting the type of annotation(s) provided.
        include_cropped: whether or not to include additional cropped samples.
        include_source_info: whether or not to include the source information of the sample.
    """
    def __init__(
        self,
        data_directory: str = "eca-data",
        data_source: DataSource = DataSource.BOTH,
        annotation_type: AnnotationType = AnnotationType.AREA,
        include_cropped: bool = True,
        include_source_info: bool = False
    ) -> None:
        super().__init__()
        self.data_directory = data_directory
        self.annotation_type = annotation_type
        self.data_source = data_source
        self.include_cropped = include_cropped
        self.include_source_info = include_source_info
        try:
            self.sample_list = self.__get_sample_list()
        except FileNotFoundError:
            raise FileNotFoundError(ERROR_MESSAGE.format(path.abspath(self.data_directory)))

    def __get_sample_list(self):
        sample_list = []

        if DataSource.CHOLEC in self.data_source:
            sample_list += get_sample_list(self.data_directory, DataSource.CHOLEC)

        if DataSource.ROBUST in self.data_source:
            sample_list += get_sample_list(self.data_directory, DataSource.ROBUST)

        if self.include_cropped:
            sample_list = add_cropped_samples(self.data_directory, sample_list)

        return sample_list

    def __len__(self):
        return len(self.sample_list)

    def __getitem__(self, key):

        if isinstance(key, slice):
            indices = range(*key.indices(len(self)))
            return (self[i] for i in indices)
            
        sample = self.sample_list[key]
        
        frame = Image.open(path.join(self.data_directory, sample['image_file']))
        frame.load()
        if self.include_cropped and sample['crop'] != None:
            frame = frame.crop(sample['crop'])
        result = (frame,)

        if AnnotationType.AREA in self.annotation_type:
            result = (*result, sample['content_area'])

        if AnnotationType.MASK in self.annotation_type:
            mask = Image.open(path.join(self.data_directory, sample['mask_file']))
            mask.load()
            if self.include_cropped and sample['crop'] != None:
                mask = mask.crop(sample['crop'])
            result = (*result, mask)

        if self.include_source_info:
            result = (*result, sample['source_info'])

        return result

# ========================================
# Some helper methods...


ERROR_MESSAGE = """
Some of the requested data cannot be found at path \"{}\"...
Please provide the correct path to the dataset, or download the dataset with the command \"download-eca\"...
"""

dataset_info = {
    DataSource.CHOLEC: {"name": "cholec-eca", "pretty_name": "CholecECA", "synapse_id": "syn32150390"},
    DataSource.ROBUST: {"name": "robust-eca", "pretty_name": "RobustECA", "synapse_id": "syn32150393"}
}

def get_sample_list(data_directory, dataset_source):
    info = dataset_info[dataset_source]
    dataset_path = path.join(data_directory, info['name'])
    with open(path.join(dataset_path, "manifest.json")) as file:
        return load(file)

def calculate_optimal_crop(circle, rectangle):

    r_w, r_h = rectangle
    c_x, c_y, c_r = circle

    aspect_ratio = r_w / r_h

    inscribed_height = 2 * (c_r - 2) / sqrt(1 + aspect_ratio * aspect_ratio)
    inscribed_width = inscribed_height * aspect_ratio

    left = max(c_x - inscribed_width / 2, 0)
    right = min(c_x + inscribed_width / 2, r_w)
    top = max(c_y - inscribed_height / 2, 0)
    bottom = min(c_y + inscribed_height / 2, r_h)

    x_scale = (right - left)
    y_scale = (bottom - top) * aspect_ratio

    scale = min(x_scale, y_scale)

    w = int(floor(scale))
    h = int(floor(scale / aspect_ratio))
    
    x = int(left + (right - left) / 2 - w / 2)
    y = int(top + (bottom - top) / 2 - h / 2)

    return x, y, x+w, y+h

def add_cropped_samples(data_directory, samples):
    new_samples = []
    for sample in samples:
        sample['crop'] = None
        new_samples.append(sample.copy())
        if sample['content_area'] != None:
            with Image.open(path.join(data_directory, sample['mask_file'])) as image:
                image_size = image.size
            sample['crop'] = calculate_optimal_crop(sample['content_area'], image_size)
            sample['content_area'] = None
            new_samples.append(sample)
    return new_samples



import numpy as np
from typing import Sequence, Optional, Tuple

class Line():
    def __init__(self, a, b) -> None:
        self.a = np.array(a)
        self.b = np.array(b)
        self.v = self.b - self.a
        self.l = np.linalg.norm(self.v)
        self.u = self.v / self.l
    
    def get_length(self):
        return self.l

    def get_point(self, t):
        return self.a + t * self.l * self.u

    def get_closest_point(self, p):
        v = p - self.a
        d = np.dot(v, self.u)
        d = max(0, min(d, self.l))
        return self.a + d * self.u

class Circle():
    def __init__(self, c, r) -> None:
        self.c = np.array(c)
        self.r = r

    def get_length(self):
        return 2 * np.pi * self.r

    def get_point(self, t):
        angle = t * 2 * np.pi
        x = np.cos(angle) * self.r
        y = np.sin(angle) * self.r
        return self.c + np.array([x, y])

    def get_closest_point(self, p):
        v = p - self.c
        u = v / np.linalg.norm(v)
        return self.c + self.r * u

class Arc():
    def __init__(self, c, r, s, e) -> None:
        self.c = np.array(c)
        self.r = r
        self.s = np.array(s)
        self.e = np.array(e)
        self.a = self.get_angle(self.e)

    def get_angle(self, p):
        v = p - self.c
        u = self.s - self.c
        angle = np.arctan2(v[1], v[0]) - np.arctan2(u[1], u[0])
        if angle < 0: angle += 2 * np.pi
        return angle
        
    def get_length(self):
        return self.a * self.r

    def get_point(self, t):
        angle = t * self.a
        v = self.s - self.c
        x = np.cos(angle) * v[0] - np.sin(angle) * v[1]
        y = np.sin(angle) * v[0] + np.cos(angle) * v[1]
        return self.c + np.array([x, y])
    
    def get_closest_point(self, p):
        angle = self.get_angle(p)
        angle = max(0, min(angle, self.a))
        return self.get_point(angle / self.a)

def make_polygon(frame_size, circle):

    h, w = frame_size

    if circle == None:
        polygon = [
            Line((0,0), (w,0)),
            Line((w,0), (w,h)),
            Line((w,h), (0,h)),
            Line((0,h), (0,0)),
        ]
        return polygon

    x, y, r = circle

    intersections = []
    is_line = []

    # Upper edge
    if y-r < 0:
        d = np.sqrt(r**2 - (y - 0)**2)
        if x-d > 0: intersections.append((x-d, 0)); is_line.append(True)
        if x+d < w: intersections.append((x+d, 0)); is_line.append(False)
        if len(is_line) == 0:
            intersections.append((w, 0)); is_line.append(True)

    if len(is_line) > 0 and is_line[-1]:
        intersections.append((w, 0)); is_line.append(True)
    
    # Right edge
    if x+r > w:
        d = np.sqrt(r**2 - (x - w)**2)
        if y-d > 0: intersections.append((w, y-d)); is_line.append(True)
        if y+d < h: intersections.append((w, y+d)); is_line.append(False)
    
    if len(is_line) > 0 and is_line[-1]:
        intersections.append((w, h)); is_line.append(True)

    # Bottom edge
    if y+r > h:
        d = np.sqrt(r**2 - (y - h)**2)
        if x+d < w: intersections.append((x+d, h)); is_line.append(True)
        if x-d > 0: intersections.append((x-d, h)); is_line.append(False)

    if len(is_line) > 0 and is_line[-1]:
        intersections.append((0, h)); is_line.append(True)

    # Left edge
    if x-r < 0:
        d = np.sqrt(r**2 - (x - 0)**2)
        if y+d < h: intersections.append((0, y+d)); is_line.append(True)
        if y-d > 0: intersections.append((0, y-d)); is_line.append(False)

    if len(is_line) > 0 and is_line[-1]:
        intersections.append((0, 0)); is_line.append(True)

    if len(intersections) == 0:
        polygon = [
            Circle((x, y), r)
        ]
        return polygon

    start = intersections
    end = intersections[1:] + intersections[:1]

    polygon = []
    for s, e, line in zip(start, end, is_line):
        if line:
            segment = Line(s, e)
        else:
            segment = Arc((x, y), r, s, e)
        polygon.append(segment)

    return polygon

def get_polygon_points(polygon, n_points):

    lengths = [seg.get_length() for seg in polygon]
    total_length = sum(lengths)

    step = total_length / n_points
    
    points = []

    for seg, length in zip(polygon, lengths):

        # No need to discretise lines, final point pair 
        # will never be in the middle of the line
        if isinstance(seg, Line):
            points.append(seg.b)
            continue

        seg_number = length // step
        for i in range(int(seg_number), 0, -1):
            point = seg.get_point(i / seg_number)
            points.append(point)

    return points

def get_smallest_dist(polygon, point):
    smallest_dist = 1e9
    other_point = None
    for seg in polygon:
        closest = seg.get_closest_point(point)
        dist = np.linalg.norm(point - closest)
        if dist < smallest_dist:
            smallest_dist = dist
            other_point = closest
    return smallest_dist, other_point

def content_area_hausdorff(
    circle_a: Optional[Sequence[int]],
    circle_b: Optional[Sequence[int]],
    frame_size: Sequence[int],
    n_points: int = 100,
    normalise: bool = True
) -> Tuple[float, Optional[Sequence[Sequence[int]]]]:
    """
    Hausdorff distance between two content areas.

    Args:
        circle_a: circle for the first content area (can be None).
        circle_b: circle for the second content area (can be None).
        frame_size: size of the image in question.
        n_points: number of points used when discretising the edges of the content areas.
        normalise: whether or not to normalise the result as if the image were 1080x1920.
    Returns:
        float: score in pixels (optionally normalised)
        tuple: the coordinates of the two points found to give the final score.
    """

    if (circle_a == circle_b):
        return 0.0, None

    polygon_a = make_polygon(frame_size, circle_a)
    polygon_b = make_polygon(frame_size, circle_b)

    points_a = get_polygon_points(polygon_a, n_points)
    points_b = get_polygon_points(polygon_b, n_points)

    hausdorff_distance = 0.0
    best_pair = None

    for point in points_b:
        dist, other_point = get_smallest_dist(polygon_a, point)
        if dist > hausdorff_distance:
            hausdorff_distance = dist
            best_pair = (point, other_point)

    for point in points_a:
        dist, other_point = get_smallest_dist(polygon_b, point)
        if dist > hausdorff_distance:
            hausdorff_distance = dist
            best_pair = (point, other_point)

    if normalise:
        hausdorff_distance = hausdorff_distance * np.linalg.norm([1080, 1920]) / np.linalg.norm(frame_size)

    return hausdorff_distance, best_pair



from setuptools import setup
import versioneer

with open("README.md") as file:
    long_description = file.read()

setup(
    name='ecadataset',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description='Python loader for the Endoscopic Content Area (ECA) dataset.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Charlie Budd',
    author_email='charles.budd@kcl.ac.uk',
    url='https://github.com/charliebudd/eca-dataset',
    license='MIT',
    package_dir={'':'src'},
    packages=['ecadataset'],
    scripts=['bin/ecadataset'],
    install_requires=['synapseclient', 'pillow', 'numpy']
)
